from neologger import Alarm, SlackNotification
from neologger.core import Condition
import time

slack = SlackNotification()
slack.set_hook("[SLACK_WEBHOOK_URL]")

alarm = Alarm(3, slack=slack)

def main():
    print("Testing NeoLogger's Alarm")

    print("Sleeping 2 Seconds")
    time_from = time.time()
    time.sleep(2)
    time_to = time.time()
    print("Awake after 2 Seconds")
    alarm.check(time_from, time_to)
    print("Sleeping 4 seconds")
    time_from = time.time()
    time.sleep(4)
    time_to = time.time()
    print("Awake after 4 seconds")
    alarm.check(time_from, time_to)

if __name__ == "__main__":
    main()


from neologger import TeamsNotification

teams = TeamsNotification()

def main():
    print("\nStarting TeamsNotification")
    print("Setting Webhook")

    teams.set_hook("YOUR_WEBHOOK_URL")

    print("Standar Notification")

    print("Setting Data")
    teams.assembly_standard_notification("NOTIFICATION'S TITLE", "This is the summary.", "This is the text within the notification body, one single paragraph.")
    print("sending Standard Notification")
    teams.send()
    result, message = teams.send()
    print("Result: " + str(result))
    print("Response: " + message)

    print("Adaptative Notification")
    print("Setting Profile Image and Name")
    teams.set_profile_image("YOUR_PUBLIC_PROFILE_IMAGE_URL")
    teams.set_profile_name("profile Name")
    print("Setting Title and ")
    teams.set_adaptative_notification("THIS IS THE NOTIFICATION TITLE")
    print("Adding data rows")
    teams.add_data("Line 1")
    teams.add_data("Line 2")
    teams.add_data("Line 3")
    print("Sending Adaptative Notification")
    result, message = teams.send()
    print("Result: " + str(result))
    print("Response: " + message)

if __name__ == "__main__":
    main()



# neologger/core/Base.py

class FontColour:
    """ANSI escape sequences for font colors."""
    BLUE = "\033[94m"         # Light blue text
    DARKBLUE = "\033[34m"     # Dark blue text
    CYAN = "\033[96m"         # Light cyan text
    DARKCYAN = "\033[36m"     # Dark cyan text
    GREEN = "\033[92m"        # Light green text
    DARKGREEN = "\033[32m"    # Dark green text
    YELLOW = "\033[93m"       # Light yellow text
    DARKYELLOW = "\033[33m"   # Dark yellow text
    RED = "\033[91m"          # Light red text
    DARKRED = "\033[31m"      # Dark red text
    MAGENTA = "\033[95m"      # Light magenta text
    DARKMAGENTA = "\033[35m"  # Dark magenta text
    GREY = "\033[37m"         # Light grey text
    DARKGREY = "\033[90m"     # Dark grey text
    BLACK = "\033[30m"        # Black text
    WHITE = "\033[97m"        # White text
    ENDC = "\033[0m"          # Reset to default text color

class BackgroundColour:
    """ANSI escape sequences for background colors."""
    BLUE = "\033[104m"         # Light blue background
    DARKBLUE = "\033[44m"      # Dark blue background
    CYAN = "\033[106m"         # Light cyan background
    DARKCYAN = "\033[46m"      # Dark cyan background
    GREEN = "\033[102m"        # Light green background
    DARKGREEN = "\033[42m"     # Dark green background
    YELLOW = "\033[103m"       # Light yellow background
    DARKYELLOW = "\033[43m"    # Dark yellow background
    RED = "\033[101m"          # Light red background
    DARKRED = "\033[41m"       # Dark red background
    MAGENTA = "\033[105m"      # Light magenta background
    DARKMAGENTA = "\033[45m"   # Dark magenta background
    GREY = "\033[47m"          # Light grey background
    DARKGREY = "\033[100m"     # Bright black background (appears as dark grey)
    BLACK = "\033[40m"         # Black background
    WHITE = "\033[107m"        # White background
    ENDC = "\033[0m"           # Reset to default background color

class FontStyle:
    """ANSI escape sequences for font styles."""
    BOLD = "\033[1m"            # Bold text
    ITALIC = "\033[3m"          # Italic text
    UNDERLINE = "\033[4m"       # Underlined text
    DOUBLEUNDERLINE = "\033[21m"  # Double underlined text
    DIM = "\033[2m"             # Dim text
    NORMAL = "\033[22m"         # Normal intensity text
    ENDC = "\033[0m"            # Reset to default text style

class Icon:
    """Unicode characters for icons."""
    MONITOR = "\U0001F5A5"     # Desktop computer icon
    DONE = "\U00002705"        # Check mark icon
    READY = "\U00002728"       # Sparkles icon
    LOCK = "\U0001F510"        # Lock with key icon
    KEY = "\U0001F511"         # Key icon
    USER = "\U0001F464"        # User silhouette icon
    ERROR = "\U000026D4"       # No entry icon
    SURPRISE = "\U00002049"    # Exclamation question mark icon
    BULLSEYE = "\U0001F3AF"    # Bullseye icon
    STOP = "\U0000270B"        # Raised hand icon
    CONFIRM = "\U00002754"     # Question mark icon
    CANCEL = "\U0001F536"      # Large red circle icon
    LOGOUT = "\U0001F463"      # Footprints icon
    CATALOGS = "\U00002699"    # Gear icon
    WARNING = "\U000026A0"     # Warning sign icon
    STAR = "\U0001F31F"        # Glowing star icon
    STOPWATCH = "\U000023F1"   # Stopwatch

class Template:
    """Predefined templates for styling."""
    NORMAL = "NORMAL"          # Normal template
    DARK = "DARK"              # Dark template
    BASE = "BASE"              # Base template

class Condition:
    BELOW = "BELOW"
    BELOW_OR_EQUAL = "BELOW_OR_EQUAL"
    EQUAL = "EQUAL"
    ABOVE = "ABOVE"
    ABOVE_OR_EQUAL = "ABOVE_OR_EQUAL"

class Letter:
    ASCII_CAPITALS = {
        "A": [
            ""
        ]
    }


from neologger import Stopwatch
import time

print("Testing NeoLogger's Stopwatch class.")


# stopwatch = Stopwatch()
# stopwatch.lap()
# time.sleep(1.25)  
# stopwatch.lap()
# time.sleep(2.55) 
# stopwatch.lap()
# time.sleep(0.35)  
# stopwatch.lap()
# time.sleep(1.87)
# stopwatch.lap()
# time.sleep(0.32)
# stopwatch.lap()
# print(stopwatch.stop())

stopwatch = Stopwatch("Tracking checkpoints using labels.")
stopwatch.lap("Starting.")
time.sleep(1.15)
stopwatch.lap("Database connection with default driver.")
time.sleep(3.25)
stopwatch.lap("Post request completed. Call to SP defaultLoggingFunction.")
time.sleep(2.3)
stopwatch.lap("Beggining to process data from Database.")
time.sleep(1.87)
stopwatch.lap("End processing data from Database.")
time.sleep(2.04)
stopwatch.lap("Another process completed.")
time.sleep(0.32)
stopwatch.lap("Function completed.")
print(stopwatch.stop())


# neologger/main.py

from datetime import datetime
import time
import logging
from .core import FontColour, BackgroundColour, FontStyle, Icon, Template, Condition
import stomp
import json
import inspect
import requests
import os

# Configure the logger
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

class NeoLogger:
    """
    A custom logger class that provides advanced logging features with customizable
    colors, styles, and templates for log messages.
    """

    def __init__(self, initialized_by, alarm=None, teams=None, slack=None, stopwatch=None):
        """
        Initialize the NeoLogger with default styling.

        Args:
            initialized_by (str): The name of the module or script that initializes the logger.
        """
        self.initialized_by = initialized_by
        self.date_colour = FontColour.WHITE
        self.date_style = FontStyle.NORMAL
        self.file_colour = FontColour.WHITE
        self.file_style = FontStyle.NORMAL
        self.function_colour = FontColour.WHITE
        self.function_style = FontStyle.NORMAL
        self.text_colour = FontColour.WHITE
        self.text_style = FontStyle.NORMAL
        self.markers = {}
        self.stopwatch = stopwatch
        self.alarm = alarm
        self.slack = slack
        self.teams = teams
        logs_location = "jsonlogs"
        logs_file = "logs.json"
        self.log_filename = logs_location + "/" + logs_file

        if not os.path.exists(logs_location):
            os.makedirs(logs_location)
        
        if os.path.isfile(self.log_filename):
            lg_file = open(logs_location + "/" + logs_file, "w")
            lg_file.close()

    def set_log_font_colour(self, date_colour, file_colour, function_colour, text_colour):
        """
        Set the font colors for different parts of the log message.

        Args:
            date_colour (str): ANSI escape code for the date color.
            file_colour (str): ANSI escape code for the file name color.
            function_colour (str): ANSI escape code for the function name color.
            text_colour (str): ANSI escape code for the log message text color.
        """
        self.date_colour = date_colour
        self.file_colour = file_colour
        self.function_colour = function_colour
        self.text_colour = text_colour

    def set_log_font_style(self, date_style, file_style, function_style, text_style):
        """
        Set the font styles for different parts of the log message.

        Args:
            date_style (str): ANSI escape code for the date style.
            file_style (str): ANSI escape code for the file name style.
            function_style (str): ANSI escape code for the function name style.
            text_style (str): ANSI escape code for the log message text style.
        """
        self.date_style = date_style
        self.file_style = file_style
        self.function_style = function_style
        self.text_style = text_style

    def set_template(self, template_name):
        """
        Apply a predefined template to the logger.

        Args:
            template_name (str): The name of the template to apply.
        """
        if template_name == Template.DARK:
            self.set_log_font_colour(
                FontColour.YELLOW, FontColour.GREEN, FontColour.CYAN, FontColour.GREY
            )
            self.set_log_font_style(
                FontStyle.BOLD, FontStyle.NORMAL, FontStyle.ITALIC, FontStyle.NORMAL
            )
        elif template_name == Template.BASE:
            self.set_log_font_colour(
                FontColour.CYAN, FontColour.YELLOW, FontColour.BLUE, FontColour.MAGENTA
            )
            self.set_log_font_style(
                FontStyle.BOLD, FontStyle.NORMAL, FontStyle.ITALIC, FontStyle.NORMAL
            )
        else:
            # Default template
            self.set_log_font_colour(
                FontColour.WHITE, FontColour.WHITE, FontColour.WHITE, FontColour.WHITE
            )
            self.set_log_font_style(
                FontStyle.NORMAL, FontStyle.NORMAL, FontStyle.NORMAL, FontStyle.NORMAL
            )

    def log_this(self, message, save_json=True):
        """
        Log a general information message with the current styling.

        Args:
            message (str): The message to log.
        """
        stack = inspect.stack()
        from_here = stack[1].function
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]  # With milliseconds
        logging.info(
            self.date_colour
            + self.date_style
            + f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        if save_json:
            self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message)
    
    def log_this_json(self, timest, source_name, function_name, log_message, log_type="INFO"):
        """
        Log a JSON message for automated processing.
        """
        log_entry = {
            "timestamp": timest,
            "type": log_type,
            "source": source_name,
            "function": function_name,
            "log": log_message,
        }

        if os.path.exists(self.log_filename) and os.path.getsize(self.log_filename) > 0:
            with open(self.log_filename, "r") as log_file:
                logs = json.load(log_file)
        else:
            logs = []

        logs.append(log_entry)

        with open(self.log_filename, "w") as log_file:
            json.dump(logs, log_file, indent=4)

    def log_this_warning(self, message):
        """
        Log a warning message with a warning icon and styling.

        Args:
            message (str): The warning message to log.
        """
        stack = inspect.stack()
        from_here = stack[1].function
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        logging.warning(
            FontColour.YELLOW
            + FontStyle.BOLD
            + "["
            + Icon.WARNING
            + " WARNING]"
            + FontColour.ENDC
            + self.date_colour
            + self.date_style
            + f" {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message, log_type="WARNING")

    def log_this_ok(self, message):
        """
        Log a success message indicating an operation was OK.

        Args:
            message (str): The success message to log.
        """
        stack = inspect.stack()
        from_here = stack[1].function
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        logging.info(
            FontColour.GREEN
            + FontStyle.BOLD
            + "["
            + Icon.DONE
            + " OK]"
            + FontColour.ENDC
            + self.date_colour
            + self.date_style
            + f" {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message)

    def log_this_error(self, message):
        """
        Log an error message with an error icon and styling.

        Args:
            message (str): The error message to log.
        """
        stack = inspect.stack()
        from_here = stack[1].function
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        logging.error(
            FontColour.RED
            + FontStyle.BOLD
            + "["
            + Icon.ERROR
            + " ERROR]"
            + FontColour.ENDC
            + self.date_colour
            + self.date_style
            + f" {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message, log_type="ERROR")

    def log_this_completed(self, message):
        """
        Log a message indicating a task has been completed.

        Args:
            message (str): The completion message to log.
        """
        stack = inspect.stack()
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        from_here = stack[1].function
        logging.info(
            FontColour.CYAN
            + FontStyle.BOLD
            + "["
            + Icon.BULLSEYE
            + " COMPLETED]"
            + FontColour.ENDC
            + self.date_colour
            + self.date_style
            + f" {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message)

    def log_this_success(self, message):
        """
        Log a success message with a star icon.

        Args:
            message (str): The success message to log.
        """
        stack = inspect.stack()
        time_stamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        from_here = stack[1].function
        logging.info(
            FontColour.MAGENTA
            + FontStyle.BOLD
            + "["
            + Icon.STAR
            + " SUCCESS]"
            + FontColour.ENDC
            + self.date_colour
            + self.date_style
            + f" {datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
        )

        self.log_this_json(time_stamp, stack[1].filename, stack[1].function, message)

    def start_stopwatch(self, title=""):
        self.stopwatch = Stopwatch(title)

    def lap(self, label=""):
        self.stopwatch.lap(label, alarm=self.alarm)

    def log_this_with_trace(self, message):
        stopwatch_data = self.stopwatch.stop()

        self.log_this(message + "\n" + stopwatch_data, save_json=False)

    def get_time_mark(self):
        """
        Get the current time in seconds since the Epoch.

        Returns:
            float: The current time in seconds.
        """
        return time.time()

    def log_with_elapsed_time(self, message, start_time):
        """
        Log a message along with the elapsed time since a given start time.

        Args:
            message (str): The message to log.
            start_time (float): The start time to calculate elapsed time from.

        Returns:
            str: A string indicating the elapsed time.
        """
        stack = inspect.stack()
        from_here = stack[1].function
        elapsed_time = time.time() - start_time
        logging.info(
            self.date_colour
            + self.date_style
            + f"{datetime.now().strftime('%Y-%m-%d %H:%M:%S')}"
            + FontStyle.ENDC
            + " - "
            + self.file_colour
            + self.file_style
            + f"{self.initialized_by} >> "
            + FontStyle.ENDC
            + self.function_colour
            + self.function_style
            + f"{from_here} | "
            + FontStyle.ENDC
            + self.text_colour
            + self.text_style
            + f"{message}"
            + FontStyle.ENDC
            + FontColour.GREY
            + FontStyle.ITALIC
            + f" [Elapsed time: {elapsed_time} seconds.]"
            + FontStyle.ENDC
        )

        if self.alarm:
            self.alarm.check(elapsed_time, start_time, summary="Function " + from_here + " raised an Alarm", title="NeoLogger Alarm.")

        return f"Elapsed time: {elapsed_time} seconds."

class StompBabbler:
    """
    A class to send messages over STOMP (Simple Text Oriented Messaging Protocol).
    """

    def __init__(self, userName, userPassword, queue, server, port):
        """
        Initialize the STOMP connection parameters.

        Args:
            userName (str): The username for the STOMP server.
            userPassword (str): The password for the STOMP server.
            queue (str): The destination queue to send messages to.
            server (str): The STOMP server address.
            port (int): The port number of the STOMP server.
        """
        self.stompUsername = userName
        self.stompPassword = userPassword
        self.stompQueue = queue
        self.stompServer = server
        self.stompPort = port

    def babble(self, message):
        """
        Send a message to the configured STOMP queue.

        Args:
            message (dict): The message payload to send.

        Returns:
            tuple: A tuple containing a boolean status and a message.
        """
        try:
            # Establish a connection to the STOMP server
            stompConnection = stomp.Connection([(self.stompServer, self.stompPort)])
            stompConnection.connect(self.stompUsername, self.stompPassword, wait=True)

            # Convert the message to JSON and send it
            json_message = json.dumps(message)
            stompConnection.send(body=json_message, destination=self.stompQueue)

            # Disconnect after sending the message
            stompConnection.disconnect()

            return True, "OK"
        except stomp.exception.ConnectFailedException as ex:
            return False, f"Connection failed: {str(ex)}"
        except Exception as ex:
            return False, str(ex)

class TeamsNotification:

    def __init__(self):
        self.webhook = ""
        self.type = "standard"
        self.title = ""
        self.theme_colour = "0076D7"
        self.summary = ""
        self.text = ""
        self.data = []
        self.teams_type = "MessageCard"
        self.profile_image = ""
        self.profile_name = ""
        self.teams_context = "http://schema.org/extensions"
        self.teams_content_type = "application/vnd.microsoft.card.adaptive"
        self.teams_schema = "http://adaptivecards.io/schemas/adaptive-card.json"
        self.hook_ready = False
        self.content_ready = False
        self.teams_version = "1.6"
    
    def set_hook(self, webhook):
        if len(webhook) > 0:
            self.webhook = webhook
            self.hook_ready = True

    def set_profile_image(self, image):
        self.profile_image = image
    
    def set_profile_name(self, name):
        self.profile_name = name
        
    def add_data(self, line):
        self.data.append({
            "type": "TextBlock",
            "text": line,
            "wrap": True
        })
    
    def assembly_teams_notification(self):
        notification_body = {}
        body = []
        body.append({"type": "TextBlock", "size": "Medium", "weight": "Bolder", "text": self.title})
        body_content = {}
        body_content["type"] = "ColumnSet"
        body_columns = []
        body_columns.append({"type": "Column", "items": [{"type": "Image", "style": "Person", "url": self.profile_image, "altText": self.profile_name, "size": "Small"}], "width": "auto"})
        body_columns.append({"type": "Column", "items": [{"type": "TextBlock", "weight": "Bolder", "text": self.profile_name, "wrap": True}, {"type": "TextBlock", "spacing": "None", "text": "Created on " + datetime.now().strftime('%d-%m-%Y at %H:%M'), "isSubtle": True, "wrap": True}], "width": "stretch"})
        body_content["columns"] = body_columns
        body.append(body_content)
        for dat in self.data:
            body.append(dat)
    
        notification_body["type"] = "AdaptiveCard"

        body_columns_data = {}
        body_columns_data["type"] = "ColumnSet"
        body_columns_data["columns"] = body_columns

        notification_body["body"] = body
        notification_body["$schema"] = self.teams_schema
        notification_body["version"] = self.teams_version

        attachments = []
        attachments_body = {}
        attachments_body["contentType"] = self.teams_content_type
        attachments_body["contentUrl"] = None
        attachments_body["content"] = notification_body
        attachments.append(attachments_body)

        notif = {}
        notif["type"] = "message"
        notif["attachments"] = attachments

        return notif
    
    def assembly_standard_notification(self, title, summary, text):
        self.type = "standard"
        self.data = []
        self.content_ready = True
        if len(title) > 0:
            self.title = title.upper()
        else:
            self.content_ready = False
        
        if len(summary) > 0:
            self.summary = summary
        else:
            self.content_ready = False

        if len(text) > 0:
            self.text = text
        else:
            self.content_ready = False
    
    def set_adaptative_notification(self, title):
        self.type = "AdaptiveCard"
        self.content_ready = True
        if len(title) > 0:
            self.title = title.upper()
        else:
            self.content_ready = False

    def send(self):
        result_ok = True
        result_message = ""

        if self.type == "standard":
            if self.content_ready == True and self.hook_ready == True:
                message = {
                    "@type": self.teams_type,
                    "@context": self.teams_context,
                    "summary": self.summary,
                    "themeColor": self.theme_colour,
                    "title": self.title,
                    "text": self.text
                }

                try:
                    response = requests.post(self.webhook, json=message)

                    if response.status_code == 200:
                        result_message = "OK"
                        self.summary = ""
                        self.title = ""
                        self.text = ""
                        self.content_ready = False
                    else:
                        result_ok = False
                        result_message = "Error" 
                except Exception as ex12:
                    result_ok = False
                    result_message = "Error: " + ex12

        elif self.type == "AdaptiveCard":
            if self.content_ready == True and self.hook_ready == True:

                message = self.assembly_teams_notification()

                try:
                    response = requests.post(self.webhook, json=message)

                    if response.status_code == 200:
                        result_message = "OK"
                        self.title = ""
                        self.text = ""
                        self.content_ready = False

                        self.data = []
                    else:
                        result_ok = False
                        result_message = "Error: " + str(response.status_code) 
                except Exception as ex12:
                    result_ok = False
                    result_message = "Error: " + ex12
        
        return result_ok, result_message


class SlackNotification:
    """
    A class to assemble and send notifications to Slack via a webhook.
    """

    def __init__(self):
        """
        Initialize the SlackNotification with default values.
        """
        self.data = []
        self.hook = ""
        self.ready = False
        self.body = None

    def add_data(self, field_name, field_value):
        """
        Add a field to the notification data.

        Args:
            field_name (str): The name of the field.
            field_value (str): The value of the field.
        """
        self.data.append({"name": field_name, "value": field_value})

    def set_hook(self, hook):
        """
        Set the Slack webhook URL.

        Args:
            hook (str): The Slack webhook URL.
        """
        self.hook = hook
        self.ready = True

    def assembly_notification(self, title, summary, icon=""):
        """
        Assemble the notification payload to send to Slack.

        Args:
            title (str): The title of the notification.
            summary (str): A summary message.
            icon (str, optional): An emoji icon to include in the header. Defaults to "".
        """
        blocks = []

        # Add a header with optional icon
        if len(icon) > 0:
            blocks.append({
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": f":{icon}: - {title}",
                    "emoji": True
                }
            })
        else:
            blocks.append({
                "type": "header",
                "text": {
                    "type": "plain_text",
                    "text": f"{title}",
                    "emoji": True
                }
            })

        blocks.append({"type": "divider"})

        # Add the summary section
        blocks.append({
            "type": "section",
            "text": {
                "type": "plain_text",
                "text": summary,
                "emoji": True
            }
        })

        # Add fields from the data list in pairs
        while self.data:
            row = {"type": "section"}
            row_fields = []

            # Pop the first item and add it to the fields
            item1 = self.data.pop()
            row_fields.append({
                "type": "mrkdwn",
                "text": f"*{item1['name']}:*\n{item1['value']}"
            })

            # Pop the second item if available and add it
            if len(self.data) > 0:
                item2 = self.data.pop()
                row_fields.append({
                    "type": "mrkdwn",
                    "text": f"*{item2['name']}:*\n{item2['value']}"
                })

            row['fields'] = row_fields
            blocks.append(row)

        blocks.append({"type": "divider"})

        # Set the assembled blocks as the message body
        self.body = {"blocks": blocks}

        self.data = []

    def send(self):
        """
        Send the assembled notification to Slack.

        Returns:
            tuple: A tuple containing a boolean status and a message.
        """
        if self.ready:
            if self.body is not None:
                try:
                    response = requests.post(
                        self.hook,
                        data=json.dumps(self.body),
                        headers={'Content-Type': 'application/json'}
                    )

                    if response.status_code == 200:
                        return True, "OK"
                    else:
                        return False, f"HTTP Error: {response.status_code}"
                except requests.exceptions.RequestException as ex:
                    return False, f"Request error: {str(ex)}"
                except Exception as ex:
                    return False, str(ex)
            else:
                return False, "Empty Body"
        else:
            return False, "Not Ready"

class DiscordNotification:
    """
    A class to assemble and send notifications to Discord via a webhook.
    Info available at: https://birdie0.github.io/discord-webhooks-guide/discord_webhook.html
    """

    class Embed:

        def __init__(self):

            self.data = {}
            self.data["title"] = "Notification"
            self.data["description"] = "Description"

        def set(self, title: str = None, description: str = None, color: str = None):

            if title:
                self.data["title"] = title
            if description:
                self.data["description"] = description
            if color:
                self.data["color"] = color

        def get(self, key):

            return self.data[key]

        def print(self):

            print(self.data)

        def add_image(self, url):

            self.data["image"] = {"url": url}

        def add_thumbnail(self, url):

            self.data["thumbnail"] = {"url": url}

        def add_author(self, name: str = None, url: str = None, icon_url: str = None):

            author = {}
            if name:
                author["name"] = name
            if url:
                author["url"] = url
            if icon_url:
                author["icon_url"] = icon_url
            if author != {}:
                self.data["author"] = author
        
        def add_field(self, name: str = None, value: str = None, inline: bool = False):

            field = {}
            if name:
                field["name"] = name
            if value:
                field["value"] = value
            if inline:
                field["inline"] = inline
            if field != {}:
                if self.data.get("fields"):
                    self.data["fields"].append(field)
                else:
                    self.data["fields"] = [field]

        def add_footer(self, text: str = None, icon_url: str = None):

            footer = {}
            if text:
                footer["text"] = text
            if icon_url:
                footer["icon_url"] = icon_url
            if footer != {}:
                self.data["footer"] = footer
                
    def __init__(self):
        """
        Initialize the DiscordNotification with default values.
        """
        self.hook = ""
        self.ready = False
        self.body = {}
        self.embeds = []

    def set_hook(self, hook):
        """
        Set the Slack webhook URL.

        Args:
            hook (str): The Discord webhook URL.
        """
        self.hook = hook
        self.ready = True

    def set_username(self, username):

        self.body["username"] = username

    def set_avatar_url(self, url):

        self.body["avatar_url"] = url

    def add_content(self, message):

        self.body["content"] = message

    def create_embed(self):

        embed = self.Embed()
        return embed

    def add_embed(self, embed):

        self.embeds.append(embed.data) 
        self.body["embeds"] = self.embeds
        self.body.pop("content", None) 

    def send(self):
        """
        Send the assembled notification to Discord.

        Returns:
            tuple: A tuple containing a boolean status and a message.
        """
        if self.ready:
            if self.body is not None:
                try:
                    response = requests.post(
                        self.hook,
                        data=json.dumps(self.body),
                        headers={'Content-Type': 'application/json'}
                    )

                    if response.status_code == 200:
                        return True, "OK"
                    else:
                        return False, f"HTTP Error: {response.status_code}"
                except requests.exceptions.RequestException as ex:
                    return False, f"Request error: {str(ex)}"
                except Exception as ex:
                    return False, str(ex)
            else:
                return False, "Empty Body"
        else:
            return False, "Not Ready"


class Table:
    """
    A class to display data in a table format.
    """

    def __init__(self):
        """
        Initialize the Table with default settings.
        """
        self.title = ""
        self.display_total = False
        self.display_border = False
        self.header = TableRow()
        self.data = []
        self.sizes = []
        self.size_format = ""
        self.table = ""
        self.table_width = 0

    def set_header(self, header):
        """
        Set the header row for the table.

        Args:
            header (list): A list of column headers.
        """
        self.header.add_data(header)

    def set_title(self, text):
        """
        Set the title of the table.

        Args:
            text (str): The title of the table.
        """
        self.title = text

    def enable_total(self):
        """Enable the display of the total number of rows."""
        self.display_total = True

    def enable_border(self):
        """Enable the display of a border around the table."""
        self.display_border = True

    def new_row(self):
        """
        Create a new row instance.

        Returns:
            TableRow: A new TableRow instance with the same number of columns as the header.
        """
        row = TableRow()
        row.total = self.header.total
        return row

    def push_row(self, row):
        """
        Add a row to the table's data.

        Args:
            row (TableRow): A TableRow instance to add to the table.
        """
        self.data.append(row)

    def from_json(self, data):
        if isinstance(data, list) and all(isinstance(item, dict) for item in data):
            try:
                all_keys = list(data[0].keys())
                
                all_rows_valid = all(set(all_keys) == set(row.keys()) for row in data)
                
                if all_rows_valid:
                    self.set_header(all_keys)  
                    
                    for dt in data:
                        row = self.new_row()
                        
                        row_content = [dt[key] for key in all_keys]
                        
                        row.fill_row(row_content)  
                        self.push_row(row) 

                    self.render()
                else:
                    self.table = "INCOMPLETED DATA"
            except TypeError:
                self.table = "INVALID JSON"
        else:
            self.table = "INVALID TYPE"
        
        return self.table

    def render(self):
        """
        Construct and print the table based on the current settings.
        """

        self.calculate_sizes()

        self.table = "\n\n"
        total_rows = 0

        # Title
        if self.title != "":
            self.table += FontStyle.BOLD + self.title.upper() + FontStyle.ENDC + "\n"
        # Border
        if self.display_border:
            self.table += "-" * self.table_width + "\n"
        # Header
        if len(self.header.rows):
            if self.display_border:
                self.table += "|" + self.size_format.format(*self.header.rows[0]) + "|\n"
            else:
                self.table += self.size_format.format(*self.header.rows[0]) + "\n"
        if self.display_border:
            self.table += "-" * self.table_width + "\n"
        # Data rows
        for dt in self.data:
            for rw in dt.rows:
                if self.display_border:
                    self.table += "|" + self.size_format.format(*rw) + "|\n"
                else:
                    self.table += self.size_format.format(*rw) + "\n"
                total_rows += 1
        # Border after rows
        if self.display_border:
            self.table += "-" * self.table_width + "\n"
        # Total row count
        if self.display_total:
            self.table += FontStyle.BOLD + FontStyle.ITALIC + "TOTAL ROWS: " + str(total_rows) + FontStyle.ENDC + "\n"
        
        self.table += "\n"
        
        return self.table

    def calculate_sizes(self):
        """
        Calculate column widths based on the longest content in each column.
        """
        max_size = 0
        self.sizes = []
        current_pos = 0
        self.size_format = ""
        self.table_width = 0

        # Calculate the max size for each column
        for hd in self.header.rows:
            for dt in hd:
                hd_col_len = len(dt)
                max_size = 0
                for rw in self.data:
                    col_len = len(str(rw.rows[0][current_pos]))
                    if col_len > max_size:
                        max_size = col_len
                    
                current_pos += 1
                
                sel_max = max_size
                if col_len > sel_max:
                    sel_max = col_len
                
                self.sizes.append(sel_max + 3)

        # Format specification
        for sz in self.sizes:
            self.size_format += "{:<" + str(sz) + "} "
            self.table_width += sz + 1
        self.table_width += 2

class TableRow:
    """
    A class to represent a row in a table.
    """

    def __init__(self):
        """Initialize the TableRow with default settings."""
        self.rows = []
        self.total = 0

    def add_data(self, data):
        """
        Add a list of data as a single row.

        Args:
            data (list): A list of values for the row.
        """
        self.total = 0
        self.rows = []
        tempo_row = []

        if len(data) > 0:
            for dt in data:
                self.total += 1
                tempo_row.append(str(dt))

        if self.total == len(data):
            self.rows.append(tempo_row)

    def fill_row(self, data):
        """
        Add additional rows of data.

        Args:
            data (list): A list of values to add as a new row.

        Returns:
            bool: True if the data length matches the header length, otherwise False.
        """
        if len(data) == self.total:
            self.rows.append(data)
            return True
        else:
            return False

class Stopwatch:
    # Initialize the Stopwatch class with an optional title
    def __init__(self, title=""):
        self.data = []  # List to store lap data (timestamps, labels, etc.)
        self.title = title  # Title for the stopwatch

    # Method to record a lap with optional label and alarm
    def lap(self, label="", alarm=None):
        # Get the current time in epoch format and a formatted timestamp
        current_epoch = time.time()
        current_timestamp = datetime.now().strftime('%d-%m-%Y %H:%M:%S.%f')[:-3]
        
        difference = 0  # Time difference between the current and previous lap
        prev_epoch = None  # To store the epoch time of the previous lap
        alarm_icon = "   "  # Default value for alarm icon (empty if no alarm)

        # Assign a default label if none is provided
        if label == "":
            label = "Mark " + str(len(self.data) + 1)
        
        # Calculate the time difference if there are previous laps
        if len(self.data) > 0:
            prev = self.data[len(self.data) - 1]  # Get the last lap's data
            prev_epoch = float(prev["EPOCH"])  # Extract the epoch time from the last lap
            difference = current_epoch - prev_epoch  # Calculate the elapsed time

            # Check the alarm if provided and handle alarm notifications
            if alarm and len(self.data) > 0:
                alarm.check(prev_epoch, current_epoch, summary=label, title="New Alarm.")
                if alarm.last_result:  # If the alarm condition is met, update the icon
                    alarm_icon = "[*]"

        # Format the epoch time and difference with specific decimal precision
        ce = "{:.3f}".format(current_epoch)  # Current epoch with 3 decimal places
        df = "{:.5f}".format(difference)  # Time difference with 5 decimal places

        # Append the new lap data to the list
        self.data.append({
            "LABEL": label,  # Label for the lap
            "TIMESTAMP": current_timestamp,  # Human-readable timestamp
            "EPOCH": ce,  # Formatted epoch time
            "ELAPSED": df + " Sec.",  # Formatted elapsed time in seconds
            "ALARM": alarm_icon  # Alarm status icon
        })

    # Method to stop the stopwatch and display the results
    def stop(self):
        # Check if there are recorded laps
        if len(self.data) > 0:
            tbl_data = Table()  # Create a Table object to display the results

            # Set the title of the results table based on the stopwatch title
            if self.title != "":
                tbl_data.set_title(Icon.STOPWATCH + " " + self.title)
            else:
                tbl_data.set_title(Icon.STOPWATCH + " RESULTS:")

            # Enable additional table settings (e.g., totals, borders)
            tbl_data.enable_total()
            tbl_data.enable_border()

            # Convert the lap data to a table format
            output = tbl_data.from_json(self.data)

            # Clear the stored lap data
            self.data = {}

            # Return the formatted results table
            return output
        else:
            # Return an empty string if no laps were recorded
            return ""


class Alarm:
    def __init__(self, threshold, condition=Condition.ABOVE_OR_EQUAL, slack=None, teams=None):
        self.threshold = threshold
        self.condition = condition
        self.slack = slack
        self.teams = teams
        self.last_result = False

    def check(self, init_time, end_time, summary="New Alarm", title="An Alarm has been raced.", icon="loud_sound"):
        evaluation = False
        self.last_result = False

        difference = end_time - init_time
        
        if self.condition == Condition.ABOVE:
            if difference > self.threshold:
                evaluation = True
        elif self.condition == Condition.ABOVE_OR_EQUAL:
            if difference >= self.threshold:
                evaluation = True
        elif self.condition == Condition.EQUAL:
            if difference == self.threshold:
                evaluation = True
        elif self.condition == Condition.BELOW:
            if difference < self.threshold:
                evaluation = True
        elif self.condition == Condition.BELOW_OR_EQUAL:
            if difference <= self.threshold:
                evaluation = True
        
        if evaluation:
            self.last_result = True
            if self.slack:
                self.slack.add_data("Threshold", str(self.threshold) + " Sec.")
                self.slack.add_data("Value", str(difference) + " Sec.")
                self.slack.assembly_notification(
                    title=title, 
                    summary=summary,
                    icon=icon
                )
                status, response = self.slack.send()

            if self.teams:
                self.teams.set_adaptative_notification("AN ALARM HAS BEEN RACED")
                self.teams.add_data("Threshold: " + str(self.threshold) + " Sec.")
                self.teams.add_data("Value: " + str(difference) + " Sec.")
                result, message = self.teams.send()


from neologger import NeoLogger, Table, Alarm, SlackNotification
from neologger.core import Template
from neologger.core import FontColour, FontStyle, Condition
import time

slack = SlackNotification()
slack.set_hook("[SLACK_WEBHOOK_URL]")
alarm = Alarm(3, slack=slack)
neologger = NeoLogger("test_neolloger.py", alarm=alarm)

def main():
    print("\nBasic example:\n")
    neologger.log_this("Starting NeoLogger")
    print("\n")

    print("\nExample with OK label:\n")
    neologger.log_this_ok("Function completed Ok.")
    print("\n")

    print("\nExample with WARNING label:\n")
    neologger.log_this_warning("Data was sent uncompleted.")
    print("\n")

    print("\nExample with COMPLETED label:\n")
    neologger.log_this_completed("Data collection stage completed.")
    print("\n")

    print("\nExample with SUCCESS label:\n")
    neologger.log_this_success("Request has been completed successfuly")
    print("\n")
    
    print("\nExample with ERROR label:\n")
    neologger.log_this_error("Something went wrong!")
    print("\n")
    
    print("\nExample with BASE Template:\n")
    neologger.set_template(Template.BASE)
    neologger.log_this("NeoLogger has been set with BASE Template")
    print("\n")

    print("\nExample with NORMAL Template:\n")
    neologger.set_template(Template.NORMAL)
    neologger.log_this("NeoLogger has been set with NORMAL Template")
    print("\n")

    print("\nExample with DARK Template:\n")
    neologger.set_template(Template.DARK)
    neologger.log_this("NeoLogger has been set with DARK Template")
    print("\n")

    print("\nExample with FontStyle customisation\n")
    neologger.set_log_font_style(FontStyle.NORMAL, FontStyle.ITALIC, FontStyle.BOLD, FontStyle.UNDERLINE)
    neologger.log_this("Font style has been customised")
    print("\n")

    neologger.set_template(Template.BASE)
    print("\nExample with Elapsed Time display\n")
    time_track = neologger.get_time_mark()
    time.sleep(4) # Adding delay
    neologger.log_with_elapsed_time("Function completed.", time_track)
    print("\n")

    print("\nExample of Table")
    table = Table()
    table.set_title("Last month sales report.")
    header = ["No", "Depto", "Name", "Top Product", "Total Sales", "Rank"]
    table.set_header(header)
    row = table.new_row()
    row_content = ["1", "IT", "Pablo Martinez", "Servers", "£12,500", "1sr"]
    row.fill_row(row_content)
    table.push_row(row)
    row = table.new_row()
    row_content = ["2", "Marketing", "Beatriz Romero", "Campain", "£11,250", "2nd"]
    row.fill_row(row_content)
    table.push_row(row)
    row = table.new_row()
    row_content = ["3", "Automotive", "Gabriela Martinez", "Peugeot 3008", "£11,108", "3rd"]
    row.fill_row(row_content)
    table.push_row(row)
    row = table.new_row()
    row_content = ["4", "Robotics", "Aurora Martinez", "Robotic Arm", "£10,090", "4th"]
    row.fill_row(row_content)
    table.push_row(row)
    table.enable_total()
    table.enable_border()
    neologger.log_this(table.render(), save_json=False)

    print("\nExample of Table from JSON")
    table = Table()
    jdata = [
        {"Id": "1011", "Code": "Ab99s0r", "Expiration": "13-10-2024", "Status": "ACTIVE", "Base": "https://endpoint.com/api/action"},
        {"Id": "1012", "Code": "12dLLd0", "Expiration": "12-11-2024", "Status": "ACTIVE", "Base": "https://endpoint.com/api/action"},
        {"Id": "1013", "Code": "5540xxD", "Expiration": "10-08-2024", "Status": "ACTIVE", "Base": "https://endpoint.com/api/action"},
        {"Id": "1014", "Code": "0dd9AsX", "Expiration": "08-05-2024", "Status": "INACTIVE", "Base": "https://endpoint.com/api/action"}
    ]
    
    table.set_title("DATA FROM JSON")
    table.enable_total()
    table.enable_border()
    neologger.log_this(table.from_json(jdata), save_json=False)

    print("Testing embedded stopwatch in neologger")
    neologger.start_stopwatch("ABC function trace")
    neologger.lap("Function is starting")
    time.sleep(0.34)
    neologger.lap("First checkpoint reached")
    time.sleep(1.26)
    neologger.lap("Second checkpoint reached")
    time.sleep(0.45)
    neologger.lap("Last checkpoint reached")
    neologger.log_this_with_trace("Function ABC has completed execution.")


if __name__ == "__main__":
    main()


# This follows the examples described in https://birdie0.github.io/discord-webhooks-guide/discord_webhook.html

from src.neologger import NeoLogger, DiscordNotification

neologger = NeoLogger("Discord tests")

DISCORD_WEBHOOK="YOUR DISCORD WEBHOOK URL"

# Basic discord notification
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
discord_notification.set_username("NeoLogger")
discord_notification.add_content("Hello world!")
sent, result = discord_notification.send()
print(sent, result)

# Testing a single, basic embed
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create embed
embed = discord_notification.create_embed()
embed.set("Alert", "Something has gone wrong", "15539236")
discord_notification.add_embed(embed)
sent, result = discord_notification.send()
print(sent, result)

# Testing a single, basic embed with an author
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create embed
embed = discord_notification.create_embed()
embed.set("Alert", "Something has gone wrong", "15539236")
embed.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
# Add embed to notification
discord_notification.add_embed(embed)
sent, result = discord_notification.send()
print(sent, result)

# Testing multiple embeds with authors
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create first embed
embed1 = discord_notification.create_embed()
embed1.set("Alert 1", "Something has gone wrong", "15539236")
embed1.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
# Create second embed
embed2 = discord_notification.create_embed()
embed2.set("Alert 2", "Something else has gone wrong", "15539236")
embed2.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
# Add embeds to notification
discord_notification.add_embed(embed1)
discord_notification.add_embed(embed2)
sent, result = discord_notification.send()
print(sent, result)

# Testing multiple embeds with authors and fields
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create first embed
embed1 = discord_notification.create_embed()
embed1.set("Alert 1", "Something has gone wrong", "15539236")
embed1.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed1.add_field("Number", "1", True)
embed1.add_field("Letter", "A", True)
# Create second embed
embed2 = discord_notification.create_embed()
embed2.set("Alert 2", "Something else has gone wrong", "15539236")
embed2.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed2.add_field("Number", "1", False)
embed2.add_field("Letter", "A", False)
# Add embeds to notification
discord_notification.add_embed(embed1)
discord_notification.add_embed(embed2)
sent, result = discord_notification.send()
print(sent, result)

# Testing multiple embedw with authors and fields and a footer
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create first embed
embed1 = discord_notification.create_embed()
embed1.set("Alert 1", "Something has gone wrong", "15539236")
embed1.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed1.add_field("Number", "1", True)
embed1.add_field("Letter", "A", True)
# Create second embed
embed2 = discord_notification.create_embed()
embed2.set("Alert 2", "Something else has gone wrong", "15539236")
embed2.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed2.add_field("Number", "1", False)
embed2.add_field("Letter", "A", False)
# Add footer to embed
embed2.add_footer("Goodbye for now!")
# Add embeds to notification
discord_notification.add_embed(embed1)
discord_notification.add_embed(embed2)
sent, result = discord_notification.send()
print(sent, result)

# Testing multiple embeds with authors, fields, a footer, thumbnails and images
discord_notification = DiscordNotification()
discord_notification.set_hook(DISCORD_WEBHOOK)
# Create first embed with thumbnail
embed1 = discord_notification.create_embed()
embed1.set("Alert 1", "Something has gone wrong", "15539236")
embed1.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed1.add_field("Number", "1", True)
embed1.add_field("Letter", "A", True)
embed1.add_thumbnail("https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
# Create second embed with image and footer
embed2 = discord_notification.create_embed()
embed2.set("Alert 2", "Something else has gone wrong", "15539236")
embed2.add_author("NeoLogger", icon_url="https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
embed2.add_field("Number", "1", False)
embed2.add_field("Letter", "A", False)
embed2.add_image("https://icons.veryicon.com/png/o/cartoon/bicolor-icon/robot-9.png")
# Add footer to embed
embed2.add_footer("Goodbye for now!")
# Add embeds to notification
discord_notification.add_embed(embed1)
discord_notification.add_embed(embed2)
sent, result = discord_notification.send()
print(sent, result)